"""
The base classes for the styling.
"""
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from six import with_metaclass

__all__ = (
    'Attrs',
    'DEFAULT_ATTRS',
    'ANSI_COLOR_NAMES',
    'Style',
    'DynamicStyle',
)


#: Style attributes.
Attrs = namedtuple('Attrs', 'color bgcolor bold underline italic blink reverse')
"""
:param color: Hexadecimal string. E.g. '000000' or Ansi color name: e.g. 'ansiblue'
:param bgcolor: Hexadecimal string. E.g. 'ffffff' or Ansi color name: e.g. 'ansired'
:param bold: Boolean
:param underline: Boolean
:param italic: Boolean
:param blink: Boolean
:param reverse: Boolean
"""

#: The default `Attrs`.
DEFAULT_ATTRS = Attrs(color=None, bgcolor=None, bold=False, underline=False,
                      italic=False, blink=False, reverse=False)


#: ``Attrs.bgcolor/fgcolor`` can be in either 'ffffff' format, or can be any of
#: the following in case we want to take colors from the 8/16 color palette.
#: Usually, in that case, the terminal application allows to configure the RGB
#: values for these names.
ANSI_COLOR_NAMES = [
    'ansiblack', 'ansiwhite', 'ansidefault',

    # Low intensity.
    'ansired', 'ansigreen', 'ansiyellow', 'ansiblue', 'ansifuchsia', 'ansiturquoise', 'ansilightgray',

    # High intensity. (Not supported everywhere.)
    'ansidarkgray', 'ansidarkred', 'ansidarkgreen', 'ansibrown', 'ansidarkblue',
    'ansipurple', 'ansiteal',
]


class Style(with_metaclass(ABCMeta, object)):
    """
    Abstract base class for prompt_toolkit styles.
    """
    @abstractmethod
    def get_attrs_for_token(self, token):
        """
        Return :class:`.Attrs` for the given token.
        """

    @abstractmethod
    def invalidation_hash(self):
        """
        Invalidation hash for the style. When this changes over time, the
        renderer knows that something in the style changed, and that everything
        has to be redrawn.
        """


class DynamicStyle(Style):
    """
    Style class that can dynamically returns an other Style.

    :param get_style: Callable that returns a :class:`.Style` instance.
    """
    def __init__(self, get_style):
        self.get_style = get_style

    def get_attrs_for_token(self, token):
        style = self.get_style()
        assert isinstance(style, Style)

        return style.get_attrs_for_token(token)

    def invalidation_hash(self):
        return self.get_style().invalidation_hash()



# Stubs for multiprocessing.managers

# NOTE: These are incomplete!

import queue
import threading
from typing import (
    Any, Callable, ContextManager, Dict, Iterable, List, Mapping, Optional,
    Sequence, Tuple, TypeVar, Union,
)

_T = TypeVar('_T')
_KT = TypeVar('_KT')
_VT = TypeVar('_VT')

class Namespace: ...

_Namespace = Namespace

class BaseManager(ContextManager[BaseManager]):
    address: Union[str, Tuple[str, int]]
    def connect(self) -> None: ...
    @classmethod
    def register(cls, typeid: str, callable: Optional[Callable] = ...,
                 proxytype: Any = ...,
                 exposed: Optional[Sequence[str]] = ...,
                 method_to_typeid: Optional[Mapping[str, str]] = ...,
                 create_method: bool = ...) -> None: ...
    def shutdown(self) -> None: ...
    def start(self, initializer: Optional[Callable[..., Any]] = ...,
              initargs: Iterable[Any] = ...) -> None: ...

class SyncManager(BaseManager):
    def BoundedSemaphore(self, value: Any = ...) -> threading.BoundedSemaphore: ...
    def Condition(self, lock: Any = ...) -> threading.Condition: ...
    def Event(self) -> threading.Event: ...
    def Lock(self) -> threading.Lock: ...
    def Namespace(self) -> _Namespace: ...
    def Queue(self, maxsize: int = ...) -> queue.Queue: ...
    def RLock(self) -> threading.RLock: ...
    def Semaphore(self, value: Any = ...) -> threading.Semaphore: ...
    def Array(self, typecode: Any, sequence: Sequence[_T]) -> Sequence[_T]: ...
    def Value(self, typecode: Any, value: _T) -> _T: ...
    def dict(self, sequence: Mapping[_KT, _VT] = ...) -> Dict[_KT, _VT]: ...
    def list(self, sequence: Sequence[_T] = ...) -> List[_T]: ...

class RemoteError(Exception): ...



#This entire module is currently experimental, because I haven't put much thought into making an elegant interface for it. Decorators are the simplest.
#This is the ultimate debugging tool when you know anything weird is happening but you don't know where.
#This is kinda like hunter.py but it doesn't hav syntax errors and it can launch a debugger
#This is the ultimate debugging tool when you know anything weird is happening but you don't know where.
#This is kinda like hunter.py but it doesn't hav syntax errors and it can launch a debugger

        
#TODO make this so that you must manually import it
anything="s;odfijspdouhpiwuerpinvcdouygw3rbsdljniuwhefhsboewbouebroscnousygeouygweubovsbpuebwpurbtpuwerb"#nobody will get tjis by coincidence lololol. debug_when(some_variable=anything) will trigger when that variable does exist
nothing="daoisdoijasodijasoijowqpnsdpvnspdnfpiwpiefbpsndpfsodfnpsodnfpsondfposndfpsodnfpsodnfpsondfpwpiw"#nobody will get tjis by coincidence lololol. debug_when(some_variable=nothing) will trigger when that variable doesn't exist
def debug_when(trap=None,**kwargs):
    #Don't use this in any code yet...the name might change into anything more palatable...
    #A decorator that goes over a tracer function(frame,event,args). This might be refined for convenience later, but this is the simplest and most versatile version.
    #Note that debug_when will slow your code down to a crawl, so use it for serious debug issues ans never ever ever in production code (unless you're writing a debugger too, in which case by all means...)
    #EXAMPLES:
    #   debug_when(i=5) #Will launch debugger if there's a variable called i with value 5
    #   debug_when(i=5,j=3) #Will launch debugger if there's a variable called i with value 5 AND theres a variable called j with value=3 (commas in the function definition represent AND)
    #   debug_when(i=anything) #Will launch debugger if there's a variable called i regardless of it's value
    #   debug_when(i=nothing) #Will launch debugger if there's NOT a variable called i, such as if it was deleted
    #   (NEEDS TO BE FIXED) UNTESTED: debug_when(anything=5) #Will launch debugger if there's any variable with value = 5
    #   (NEEDS TO BE FIXED) UNTESTED: debug_when(nothing=5) #Will launch debugger if there's NOT any variable with value = 5
    #   (NEEDS TO BE FIXED) UNTESTED: debug_when(anything=lambda x:x>5) #Will launch debugger if there's any variable with value greater than 5
    #   (NEEDS TO BE FIXED) UNTESTED: debug_when(nothing=lambda x:x>5) #Will launch debugger if there's NOT any variable with value greater than 5
    #   debug_when(i=lambda i:i>5) #Will launch debugger if there's a variable called i which has value greater than 5
    #   debug_when(lambda frame,event,args:'cv2' in sys.modules)#Debugger when cv2 is imported
    #   @debug_when#Can also be used as a decorator
    #   def _(frame,event,args):
    #       return 'hello world' in frame.f_locals.values() #Will launch debuger when there's a variable (regardless of name) with value 'hello world'

    import sys,pudb

    def set_pudb_trace_at_frame(frame,event,args):
        #This method is still hidden because I'm not confident in it yet
        import pudb
        dbg = pudb._get_debugger()
        dbg.set_trace(frame)
        #dbg.break_here(frame)
        #dbg.stop_here(frame)
        dbg.trace_dispatch(frame, event, args)

    if not callable(trap):
        #Syntactic sugar for debug_when's trap
        def trap(frame,event,args):
            #EXAMPLE: debug_when(i=5,j=6,x=lambda:x>5)#Will launch debugger if there's a variable called i==5 and a variable called j==6
            for name,check in kwargs.items():
                name_exists=name in frame.f_locals
                if name=='anything' or name=='nothing':
                    flag=False#If some value in the namespace satisfies the check
                    for value in frame.f_locals.values():
                        if callable(check):
                            if check(value):
                                flag=True
                                break
                        elif check==value or check is anything:
                            flag=True
                            break
                    if    name=='nothing'  and flag==True \
                       or flag=='anything' and flag==False:
                        return False
                elif not name_exists and check is not nothing:
                    return False
                elif check is nothing:
                    return False
                else:
                    value=frame.f_locals[name]
                    if callable(check):
                        if not check(value):
                            return False
                    elif check!=value and check is not anything:#If we set check to None, it will break if the variable exists at all
                        return False
            return True

    def tracer(frame,event,args):
        if trap(frame,event,args):
            set_pudb_trace_at_frame(frame,event,args)
        return tracer#We must return the tracer otherwise python will lose track of it

    sys.settrace(tracer)

def debug_when_namespace_satisfies(condition):
    #This is a decorator function
    #Namespace will be passed to condition as kwargs
    @debug_when
    def _(frame,event,args):
        namespace=frame.f_locals
        return condition(**namespace)

def theres_a_value_in_namespace_such_that(condition):
    def trap(frame,event,args):
        for key,value in frame.f_locals.items():
            if condition(value):
                return True
        return False
    return trap

def theres_a_value_equal_to(value,equals=lambda x,y:x==y):
    return theres_a_value_in_namespace_such_that(lambda x:equals(x,value))

if __name__=='__main__':
    from rp import *
    def f():
        i=0
        for _ in range(100) :
            print('i=',i)
            i+=1
            i%=30
    #@debug_when#the more robust but harder to read/write way: use debug_when as a decorator
    #def _(frame,event,args):
        #return 'i' in frame.f_locals and frame.f_locals['i']==15

    # debug_when(theres_a_value_equal_to(14))

    # debug_when(i=13)

    debug_when(i=lambda i:i>13)
    f()


from jinja2.environment import Environment as Environment, Template as Template
from jinja2.loaders import BaseLoader as BaseLoader, FileSystemLoader as FileSystemLoader, PackageLoader as PackageLoader, DictLoader as DictLoader, FunctionLoader as FunctionLoader, PrefixLoader as PrefixLoader, ChoiceLoader as ChoiceLoader, ModuleLoader as ModuleLoader
from jinja2.bccache import BytecodeCache as BytecodeCache, FileSystemBytecodeCache as FileSystemBytecodeCache, MemcachedBytecodeCache as MemcachedBytecodeCache
from jinja2.runtime import Undefined as Undefined, DebugUndefined as DebugUndefined, StrictUndefined as StrictUndefined, make_logging_undefined as make_logging_undefined
from jinja2.exceptions import TemplateError as TemplateError, UndefinedError as UndefinedError, TemplateNotFound as TemplateNotFound, TemplatesNotFound as TemplatesNotFound, TemplateSyntaxError as TemplateSyntaxError, TemplateAssertionError as TemplateAssertionError
from jinja2.filters import environmentfilter as environmentfilter, contextfilter as contextfilter, evalcontextfilter as evalcontextfilter
from jinja2.utils import Markup as Markup, escape as escape, clear_caches as clear_caches, environmentfunction as environmentfunction, evalcontextfunction as evalcontextfunction, contextfunction as contextfunction, is_undefined as is_undefined, select_autoescape as select_autoescape



# Stubs for six.moves.urllib.parse
from urlparse import ParseResult as ParseResult
from urlparse import SplitResult as SplitResult
from urlparse import parse_qs as parse_qs
from urlparse import parse_qsl as parse_qsl
from urlparse import urldefrag as urldefrag
from urlparse import urljoin as urljoin
from urlparse import urlparse as urlparse
from urlparse import urlsplit as urlsplit
from urlparse import urlunparse as urlunparse
from urlparse import urlunsplit as urlunsplit
from urllib import quote as quote
from urllib import quote_plus as quote_plus
from urllib import unquote as unquote
from urllib import unquote_plus as unquote_plus
from urllib import urlencode as urlencode
from urllib import splitquery as splitquery
from urllib import splittag as splittag
from urllib import splituser as splituser
from urlparse import uses_fragment as uses_fragment
from urlparse import uses_netloc as uses_netloc
from urlparse import uses_params as uses_params
from urlparse import uses_query as uses_query
from urlparse import uses_relative as uses_relative



from __future__ import unicode_literals
from rp.prompt_toolkit.document import Document
from rp.prompt_toolkit.enums import DEFAULT_BUFFER, DUMMY_BUFFER
from rp.prompt_toolkit.filters import HasFocus, Condition
from rp.prompt_toolkit.key_binding.vi_state import InputMode
from rp.prompt_toolkit.keys import Keys

__all__ = (
    'load_custom_pdb_key_bindings',
)


def load_custom_pdb_key_bindings(ptpdb, registry):
    """
    Custom key bindings.
    """
    handle = registry.add_binding

    source_code_has_focus = HasFocus('source_code') & Condition(
        lambda cli: not ptpdb.python_input.show_exit_confirmation)

    def return_text(event, text):
        buffer = event.cli.buffers[DEFAULT_BUFFER]
        buffer.document = Document(text)
        event.cli.set_return_value(buffer.document)

    @handle(Keys.ControlX, eager=True)
    def _(event):
        """
        Switch focus between source code and CLI.

        Eager, because we want to ignore CtrlX-CtrlE and other key bindings
        starting with CtrlX.
        """
        # vi_state = ptpdb.python_input.key_bindings_manager.get_vi_state(event.cli)

        if event.cli.current_buffer_name == DEFAULT_BUFFER:
            event.cli.focus('source_code')
            # vi_state.input_mode = InputMode.NAVIGATION

        elif event.cli.current_buffer_name == 'source_code' and not ptpdb.callstack_focussed:
            ptpdb.callstack_focussed = True
            event.cli.focus(DUMMY_BUFFER)

        else:
            ptpdb.callstack_focussed = False
            event.cli.focus(DEFAULT_BUFFER)
            # vi_state.input_mode = InputMode.INSERT

    @handle(' ', filter=source_code_has_focus)
    @handle('b', filter=source_code_has_focus)
    @handle(Keys.ControlJ, filter=source_code_has_focus)
    def _(event):
        """
        Set/clear break.
        """
        lineno = event.cli.current_buffer.document.cursor_position_row + 1

        filename = ptpdb.canonic(ptpdb.curframe.f_code.co_filename)
        breaks = ptpdb.breaks

        if lineno in breaks.get(filename, []):
            ptpdb.clear_break(filename, lineno)
        else:
            ptpdb.set_break(filename, lineno)

    @handle('n', filter=source_code_has_focus)
    def _(event):
        """
        Debug: Next.
        """
        return_text(event, 'next')

    @handle('s', filter=source_code_has_focus)
    def _(event):
        """
        Debug: Step
        """
        return_text(event, 'step')

    @handle('c', filter=source_code_has_focus)
    def _(event):
        """
        Debug: Continue.
        """
        return_text(event, 'continue')

    @handle('q', filter=source_code_has_focus)
    def _(event):
        " Quit. "
        ptpdb.python_input.show_exit_confirmation = True

    @handle(Keys.ControlC, filter=~HasFocus(DEFAULT_BUFFER))
    def _(event):
        " Focus prompt again, wherever we are. "
        # vi_state = ptpdb.python_input.key_bindings_manager.get_vi_state(event.cli)

        ptpdb.callstack_focussed = False
        event.cli.focus(DEFAULT_BUFFER)
        # vi_state.input_mode = InputMode.INSERT

    # Call stack key bindings.

    call_stack_has_focus = Condition(lambda cli: ptpdb.callstack_focussed)
    handle = registry.add_binding

    @handle(Keys.Up, filter=call_stack_has_focus)
    @handle(Keys.ControlP, filter=call_stack_has_focus)
    @handle('k', filter=call_stack_has_focus)
    def _(event):
        " Go to previous frame. "
        if ptpdb.callstack_selected_frame > 0:
            ptpdb.callstack_selected_frame -= 1

    @handle(Keys.Down, filter=call_stack_has_focus)
    @handle(Keys.ControlN, filter=call_stack_has_focus)
    @handle('j', filter=call_stack_has_focus)
    def _(event):
        " Go to next frame. "
        if ptpdb.callstack_selected_frame < len(ptpdb.stack) - 1:
            ptpdb.callstack_selected_frame += 1

    @handle(Keys.ControlJ, filter=call_stack_has_focus)
    def _(event):
        """
        Go up/down to the selected frame.
        """
        # Find index of current frame.
        current = 0
        selected = ptpdb.callstack_selected_frame

        for i, (frame, lineno) in enumerate(ptpdb.stack):
            if frame is ptpdb.curframe:
                current = i

        if current > selected:
            return_text(event, 'up %i' % (current - selected))

        elif current < selected:
            return_text(event, 'down  %i' % (selected - current))



# Stubs for _subprocess

# NOTE: These are incomplete!

from typing import Mapping, Any, Tuple

CREATE_NEW_CONSOLE = 0
CREATE_NEW_PROCESS_GROUP = 0
STD_INPUT_HANDLE = 0
STD_OUTPUT_HANDLE = 0
STD_ERROR_HANDLE = 0
SW_HIDE = 0
STARTF_USESTDHANDLES = 0
STARTF_USESHOWWINDOW = 0
INFINITE = 0
DUPLICATE_SAME_ACCESS = 0
WAIT_OBJECT_0 = 0

# TODO not exported by the Python module
class Handle:
    def Close(self) -> None: ...

def GetVersion() -> int: ...
def GetExitCodeProcess(handle: Handle) -> int: ...
def WaitForSingleObject(handle: Handle, timeout: int) -> int: ...
def CreateProcess(executable: str, cmd_line: str,
                  proc_attrs, thread_attrs,
                  inherit: int, flags: int,
                  env_mapping: Mapping[str, str],
                  curdir: str,
                  startupinfo: Any) -> Tuple[Any, Handle, int, int]: ...
def GetModuleFileName(module: int) -> str: ...
def GetCurrentProcess() -> Handle: ...
def DuplicateHandle(source_proc: Handle, source: Handle, target_proc: Handle,
                    target: Any, access: int, inherit: int) -> int: ...
def CreatePipe(pipe_attrs, size: int) -> Tuple[Handle, Handle]: ...
def GetStdHandle(arg: int) -> int: ...
def TerminateProcess(handle: Handle, exit_code: int) -> None: ...



from __future__ import unicode_literals

from rp.prompt_toolkit.cache import FastDictCache
from rp.prompt_toolkit.token import Token
from rp.prompt_toolkit.utils import get_cwidth

from collections import defaultdict, namedtuple

__all__ = (
    'Point',
    'Size',
    'Screen',
    'Char',
)


Point = namedtuple('Point', 'y x')
Size = namedtuple('Size', 'rows columns')


class Char(object):
    """
    Represent a single character in a :class:`.Screen`.

    This should be considered immutable.
    """
    __slots__ = ('char', 'token', 'width')

    # If we end up having one of these special control sequences in the input string,
    # we should display them as follows:
    # Usually this happens after a "quoted insert".
    display_mappings = {
        '\x00': '^@',  # Control space
        '\x01': '^A',
        '\x02': '^B',
        '\x03': '^C',
        '\x04': '^D',
        '\x05': '^E',
        '\x06': '^F',
        '\x07': '^G',
        '\x08': '^H',
        '\x09': '^I',
        '\x0a': '^J',
        '\x0b': '^K',
        '\x0c': '^L',
        '\x0d': '^M',
        '\x0e': '^N',
        '\x0f': '^O',
        '\x10': '^P',
        '\x11': '^Q',
        '\x12': '^R',
        '\x13': '^S',
        '\x14': '^T',
        '\x15': '^U',
        '\x16': '^V',
        '\x17': '^W',
        '\x18': '^X',
        '\x19': '^Y',
        '\x1a': '^Z',
        '\x1b': '^[',  # Escape
        '\x1c': '^\\',
        '\x1d': '^]',
        '\x1f': '^_',
        '\x7f': '^?',  # Backspace
    }

    def __init__(self, char=' ', token=Token):
        # If this character has to be displayed otherwise, take that one.
        char = self.display_mappings.get(char, char)

        self.char = char
        self.token = token

        # Calculate width. (We always need this, so better to store it directly
        # as a member for performance.)
        self.width = get_cwidth(char)

    def __eq__(self, other):
        return self.char == other.char and self.token == other.token

    def __ne__(self, other):
        # Not equal: We don't do `not char.__eq__` here, because of the
        # performance of calling yet another function.
        return self.char != other.char or self.token != other.token

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.char, self.token)


_CHAR_CACHE = FastDictCache(Char, size=1000 * 1000)
Transparent = Token.Transparent


class Screen(object):
    """
    Two dimentional buffer of :class:`.Char` instances.
    """
    def __init__(self, default_char=None, initial_width=0, initial_height=0):
        if default_char is None:
            default_char = _CHAR_CACHE[' ', Transparent]

        self.data_buffer = defaultdict(lambda: defaultdict(lambda: default_char))

        #: Escape sequences to be injected.
        self.zero_width_escapes = defaultdict(lambda: defaultdict(lambda: ''))

        #: Position of the cursor.
        self.cursor_position = Point(y=0, x=0)

        #: Visibility of the cursor.
        self.show_cursor = True

        #: (Optional) Where to position the menu. E.g. at the start of a completion.
        #: (We can't use the cursor position, because we don't want the
        #: completion menu to change its position when we browse through all the
        #: completions.)
        self.menu_position = None

        #: Currently used width/height of the screen. This will increase when
        #: data is written to the screen.
        self.width = initial_width or 0
        self.height = initial_height or 0

    def replace_all_tokens(self, token):
        """
        For all the characters in the screen. Set the token to the given `token`.
        """
        b = self.data_buffer

        for y, row in b.items():
            for x, char in row.items():
                b[y][x] = _CHAR_CACHE[char.char, token]


class WritePosition(object):
    def __init__(self, xpos, ypos, width, height, extended_height=None):
        assert height >= 0
        assert extended_height is None or extended_height >= 0
        assert width >= 0
        # xpos and ypos can be negative. (A float can be partially visible.)

        self.xpos = xpos
        self.ypos = ypos
        self.width = width
        self.height = height
        self.extended_height = extended_height or height

    def __repr__(self):
        return '%s(%r, %r, %r, %r, %r)' % (
            self.__class__.__name__,
            self.xpos, self.ypos, self.width, self.height, self.extended_height)
